;
; Purpose:
;	To make a simple interrupt-driven node that implements S.N.A.P in AVR-assembler.
;
; Author:
;	Gran Nordahl, e99_gno@e.kth.se
;
; Credits:
;	High Tech Horizon, http://www.hth.com/ [S.N.A.P]
;	John C. Wren [CRC16]
;
; File, Version, Date:
;	http://www.e.kth.se/~e99_gno/snap/snap02.asm, 1.0, 2000-09-16
;
; Target:
;	AT90S8515/STK200, but will probably work on any AVR with an UART.
;
; Description:
;	DB1 of the received packet is shown on the LEDs of STK200 (PORTB).
;	ACK is sent with the accumulated number of ACKs in DB1
;
; Accepted packet structure:
;	HDB2: 1-3 bytes destination/source adress, No flags, ACK request.
;	HDB1: Command mode not supported, CRC16, 1 data byte.
;
; Support:
;	Ask kindly
;
; Terms of use:
;	For non-commercial purposes you may use this code as you want.
;
;
; For more information about S.N.A.P, go to http://www.hth.com/snap/
;






.include "8515def.inc"

.def	ACK_COUNT = r12		;Counter for ACKs
.def	SNAP	= r13		;Offset in pointertable
.def	CRC_H	= r14		;Current CRC-value, 8 upper bits
.def	CRC_L	= r15		;Current CRC-value, 8 lower bits
.def	TEMP	= r16		;General purpose register
.def	COUNT	= r17		;Used in sub CRC16
.def	DATA	= r18		;Stores data from the UART and is used in sub CRC16
.def	HDB2	= r20		;Stores HDB2 of incoming packet
.def	HDB1	= r21		;Stores HDB1 of incoming packet
.def	SAB3	= r22		;Stores SAB3 of incoming packet. Used later as DAB3.
.def	SAB2	= r23		;Stores SAB2 of incoming packet. Used later as DAB2.
.def	SAB1	= r24		;Stores SAB1 of incoming packet. Used later as DAB1.
.def	DB1	= r25		;Stores DB1  of incoming packet.

.set	ME	= 0x000002	;Nodes adress, up to 24 bits!
.set	SYNC	= 0b01010100	;Syncronisation byte

;Accepted packet structure:
.set	RX_HDB2	= 0b11110001	;3 bytes destination/source adress, No flags, ACK request.
.set	RX_HDB1	= 0b01000001	;Command mode not supported, CRC16, 1 Data byte.






;Interrupt Vector
	rjmp	RESET		;Reset Handler
.org	URXCaddr
	rjmp	UART_RXC_TXC	;UART RX Complete Handler
.org	UTXCaddr	
	rjmp	UART_RXC_TXC	;UART TX Complete Handler






RESET:

;Init STACK
	ldi	TEMP, low(RAMEND)
	out	SPL, TEMP
	ldi	TEMP, high(RAMEND)
	out	SPH, TEMP

;Init PORTB
	ser	TEMP
	out	DDRB, TEMP	;PORTB = output
	out	PORTB, TEMP	;All LEDs cleared (LEDs inverted on STK200)

;Init ACK_COUNT
	clr	ACK_COUNT
		
;Init CRC16 (initial remainder = 0x0000)
	clr	CRC_H
	clr	CRC_L
	
;Init S.N.A.P related thing
	clr	SNAP		;Offset in pointertable = 0 ==> no SYNC / new packet

;Init UART
	ldi	TEMP, 103	;103 -> 2400bps, 51 -> 4800bps,
	out	UBRR, TEMP	; 25 -> 9600bps, 12 -> 19200bps @ 4MHz
	ldi	TEMP, 0b10010000;RX Complete Interrupt & Receiver enabled
	out	UCR, TEMP

;Enable interrupts
	sei






;**********************************************************************************
;This is an interrupt-driven node, so please make some funny code and put it here!
;Reception and transmission of S.N.A.P-packets are taken care of in the background!
;Tjohooooo!!! (Swedish for "Whoopeeeee!!!")
MAIN:	rjmp	MAIN		;Eternal loop
;**********************************************************************************






;*** UART RX & TX Complete Handlers ***
UART_RXC_TXC:

;Disable interrupts
	cli
	
;Prepares an Indirect Jump [ijmp] to current position in decode-cycle
	push	ZH
	push	ZL
	push	TEMP
	push	DATA
	push	COUNT
	in	DATA, UDR
	ldi	ZH, high(SNAP0*2)	;Get index to pointertable
	ldi	ZL, low(SNAP0*2)
	clc
	clr	TEMP
	lsl	SNAP			;Offset * 2 ==> word-offset
	rol	TEMP
	add	ZL, SNAP
	adc	ZH, TEMP		;Offset in pointertable added to index
	clc
	lsr	TEMP
	ror	SNAP			;Offset restored
	lpm
	mov	TEMP, r0
	adiw	ZL, 1
	lpm				;Pointer for ijmp fetched from pointertable
	mov	ZH, r0
	mov	ZL, TEMP
	ijmp				;Indirect Jump to some place below. The pointertable
					;tells you where to look for the next instruction.






;End of UART RX & TX Complete Handlers
NO_SYNC:				;Reset decode-cycle and related stuff
	sbi	USR, TXC		;Ensures that we'll get correct TX-interrupt
	ldi	TEMP, 0b10010000	;RX Complete Interrupt & Receiver enabled
	out	UCR, TEMP		;TX Complete Interrupt & Transmitter disabled
	clr	CRC_H
	clr	CRC_L
	clr	SNAP
	rjmp	UART_RXC_TXC_END
CRC16_INC_SNAP:
	rcall	CRC16			;Calculate CRC16
INC_SNAP:
	inc	SNAP			;Move to next position in pointertable
UART_RXC_TXC_END:
	pop	COUNT
	pop	DATA
	pop	TEMP
	pop	ZL
	pop	ZH
	sei				;Enable interrupts
	reti				;End of interrupt routine






R_SYNC:	cpi	DATA, SYNC		;SYNC?
	brne	NO_SYNC			;No:	Skip packet
	rjmp	INC_SNAP		;Yes:	Move to position for R_HDB2 in pointertable



R_HDB2:	mov	HDB2, DATA		;Stores HDB2
	sbr	DATA, 0b11110000	;This allows 1-3 DAB & SAB
	cpi	DATA, RX_HDB2		;HDB2 OK?
	brne	NO_SYNC			;No:	Skip packet
	mov	DATA, HDB2		;Yes:	Calculate CRC16 and
	rjmp	CRC16_INC_SNAP		;	move to position for R_HDB1 in pointertable



R_HDB1:	mov	HDB1, DATA		;Stores HDB1
	cpi	DATA, RX_HDB1		;HDB1 OK?
	brne	NO_SYNC			;No:	Skip packet
	rcall	CRC16			;Yes:	Calculate CRC16



;1-3 DAB?:
	mov	TEMP, HDB2
	cbr	TEMP, 0b00111111
	inc	SNAP			;Move to position for R_DAB3 in pointertable
	cpi	TEMP, 0b11000000	;3 DAB?
	breq	UART_RXC_TXC_END	;Yes
	inc	SNAP			;No: Move to position for R_DAB2 in pointertable
	cpi	TEMP, 0b10000000	;2 DAB?
	breq	UART_RXC_TXC_END	;Yes
	inc	SNAP			;No: Move to position for R_DAB1 in pointertable
	cpi	TEMP, 0b01000000	;1 DAB?
	breq	UART_RXC_TXC_END	;Yes
	rjmp	NO_SYNC			;No: 0 DAB = error ==> skip packet






;Destination Adress = ME?	
R_DAB3:	cpi	DATA, byte3(ME)		;DAB3 = bit 23-16 of ME?
	brne	NO_SYNC			;No:	Skip packet
	rjmp	CRC16_INC_SNAP		;Yes:	Calculate CRC16 and
					;	move to position for R_DAB2 in pointertable


R_DAB2:	cpi	DATA, high(ME)		;DAB2 = bit 15-8 of ME?
	brne	NO_SYNC			;No:	Skip packet
	rjmp	CRC16_INC_SNAP		;Yes:	Calculate CRC16 and
					;	move to position for R_DAB1 in pointertable

	
R_DAB1:	cpi	DATA, low(ME)		;DAB1 = bit 7-0 of ME?
	brne	NO_SYNC			;No:	Skip packet
	rcall	CRC16			;Yes:	Calculate CRC16



;1-3 SAB?:
	mov	TEMP, HDB2
	cbr	TEMP, 0b11001111
	inc	SNAP			;Move to position for R_SAB3 in pointertable
	cpi	TEMP, 0b00110000	;3 SAB?
	breq	UART_RXC_TXC_END	;Yes
	inc	SNAP			;No: Move to position for R_SAB2 in pointertable
	cpi	TEMP, 0b00100000	;2 SAB?
	breq	UART_RXC_TXC_END	;Yes
	inc	SNAP			;No: Move to position for R_SAB1 in pointertable
	cpi	TEMP, 0b00010000	;1 SAB?
	breq	UART_RXC_TXC_END	;Yes
	rjmp	NO_SYNC			;No: 0 SAB = error ==> skip packet






;Saves Source Adress [SABx] to use as Destination Adress [DABx] in ACK/NAK-packet	
R_SAB3:	mov	SAB3, DATA		;Store SAB3
	rjmp	CRC16_INC_SNAP		;Calculate CRC16 and
					;move to position for R_SAB2 in pointertable


R_SAB2:	mov	SAB2, DATA		;Store SAB2
	rjmp	CRC16_INC_SNAP		;Calculate CRC16 and
					;move to position for R_SAB1 in pointertable

	
R_SAB1:	mov	SAB1, DATA		;Store SAB1
	rjmp	CRC16_INC_SNAP		;Calculate CRC16 and
					;move to position for R_DB1 in pointertable


R_DB1:	mov	DB1, DATA		;Stores DB1
	rjmp	CRC16_INC_SNAP		;Calculate CRC16 and
					;move to position for R_CRC2 in pointertable


R_CRC2:	rcall	CRC16			;Calculate CRC16
	rjmp	INC_SNAP		;Move to position for R_CRC1 in pointertable



R_CRC1:	rcall	CRC16			;Calculate CRC16

;Prepares HDB2 & HDB1:
	mov	TEMP, HDB2		;The following swaps DD and SS of HDB2
	cbr	HDB2, 0b00111111
	lsr	HDB2
	lsr	HDB2
	cbr	TEMP, 0b11001111
	lsl	TEMP
	lsl	TEMP	
	add	HDB2, TEMP		;DD and SS of HDB2 swapped!

	sbr	HDB2, 0b00000010	;ACK response
	tst	CRC_H
	brne	T_NAK			;If CRC16 != 0, then create a NAK-packet
	tst	CRC_L
	breq	T_ACK			;If CRC16 != 0, then create a NAK-packet
T_NAK:	sbr	HDB2, 0b00000001	;NAK response
	ldi	HDB1, 0b01000000	;Command mode not supported, CRC16, 0 Data bytes
	clr	CRC_H
	clr	CRC_L
	rjmp	T_SYNC

T_ACK:	ldi	HDB1, 0b01000001	;Command mode not supported, CRC16, 1 Data byte
	inc	ACK_COUNT






;******************************************************************************
;Perhaps a decoder for data that came with the packet is more useful than this.
;So: Please make some useful code and put it here!
;******************************************************************************
; Shows DB1 on LEDs
	com	DB1
	out	PORTB, DB1	;Inverted DB1 to LEDs (LEDs on STK200 are inverted)






T_SYNC:	sbi	USR, TXC		;Ensures that we'll get correct TX-interrupt
	ldi	TEMP, 0b01001000	;TX Complete Interrupt & Transmitter enabled
	out	UCR, TEMP		;RX Complete Interrupt & Receiver disabled
	ldi	DATA, SYNC
	out	UDR, DATA		;Sends SYNC
	rjmp	INC_SNAP		;Move to position for T_HDB2 in pointertable


	
T_HDB2:	mov	DATA, HDB2		;Sub CRC16 always called with data in DATA
	out	UDR, DATA		;Sends HDB2
	rjmp	CRC16_INC_SNAP		;Calculate CRC16 and
					;move to position for T_HDB1 in pointertable


T_HDB1:	mov	DATA, HDB1		;Sub CRC16 always called with data in DATA
	out	UDR, DATA		;Sends HDB1
	rcall	CRC16			;Calculate CRC16






;1-3 DAB?:
	mov	TEMP, HDB2
	cbr	TEMP, 0b00111111
	inc	SNAP			;Move to position for T_DAB3 in pointertable
	cpi	TEMP, 0b11000000	;3 DAB?
	breq	JMP2_UART_RXC_TXC_END	;Yes
	inc	SNAP			;No: Move to position for T_DAB2 in pointertable
	cpi	TEMP, 0b10000000	;2 DAB?
	breq	JMP2_UART_RXC_TXC_END	;Yes
	inc	SNAP			;No: Move to position for T_DAB1 in pointertable
	cpi	TEMP, 0b01000000	;1 DAB?
	breq	JMP2_UART_RXC_TXC_END	;Yes
	
	;If this happens we got an evil error!
	rjmp	NO_SYNC			;No: 0 SAB = error ==> skip packet






T_DAB3:	mov	DATA, SAB3		;Sub CRC16 always called with data in DATA
	out	UDR, DATA		;Sends SAB3 of received packet as DAB3
	rjmp	CRC16_INC_SNAP		;Calculate CRC16 and
					;move to position for T_DAB2 in pointertable


T_DAB2:	mov	DATA, SAB2		;Sub CRC16 always called with data in DATA
	out	UDR, DATA		;Sends SAB2 of received packet as DAB2
	rjmp	CRC16_INC_SNAP		;Calculate CRC16 and
					;move to position for T_DAB1 in pointertable


T_DAB1:	mov	DATA, SAB1		;Sub CRC16 always called with data in DATA
	out	UDR, DATA		;Sends SAB1 of received packet as DAB1
	rcall	CRC16			;Calculate CRC16



;1-3 SAB?:
	mov	TEMP, HDB2
	cbr	TEMP, 0b11001111
	inc	SNAP			;Move to position for T_SAB3 in pointertable
	cpi	TEMP, 0b00110000	;3 SAB?
	breq	JMP2_UART_RXC_TXC_END	;Yes
	inc	SNAP			;No: Move to position for T_SAB2 in pointertable
	cpi	TEMP, 0b00100000	;2 SAB?
	breq	JMP2_UART_RXC_TXC_END	;Yes
	inc	SNAP			;No: Move to position for T_SAB1 in pointertable
	cpi	TEMP, 0b00010000	;1 SAB?
	breq	JMP2_UART_RXC_TXC_END	;Yes

	;If this happens we got an evil error!
	rjmp	NO_SYNC			;No: 0 SAB = error ==> skip packet






T_SAB3:	ldi	DATA, byte3(ME)		;Sub CRC16 always called with data in DATA
	out	UDR, DATA		;Sends bit 23-16 of ME as SAB3
	rjmp	CRC16_INC_SNAP		;Calculate CRC16 and
					;move to position for T_DAB2 in pointertable


T_SAB2:	ldi	DATA, high(ME)		;Sub CRC16 always called with data in DATA
	out	UDR, DATA		;Sends bit 15-8 of ME as SAB2
	rjmp	CRC16_INC_SNAP		;Calculate CRC16 and
					;move to position for T_DAB1 in pointertable


T_SAB1:	ldi	DATA, low(ME)		;Sub CRC16 always called with data in DATA
	out	UDR, DATA		;Sends bit 7-0 of ME as SAB1
	rcall	CRC16			;Calculate CRC16
	inc	SNAP			;Move to position for T_DB1 in pointertable
	cbr	HDB1, 0b11110000	;1 or 0 Data Bytes?
	brne	T_DB0			;1
	inc	SNAP			;0: Move to position for T_CRC2 in pointertable
T_DB0:	rjmp	UART_RXC_TXC_END



T_DB1:	mov	DATA, ACK_COUNT		;Sub CRC16 always called with data in DATA
	out	UDR, DATA		;ACK_COUNT as DB1
	rjmp	CRC16_INC_SNAP		;Calculate CRC16 and
					;move to position for T_CRC2 in pointertable


T_CRC2:	mov	DATA, CRC_H		;Sub CRC16 always called with data in DATA
	out	UDR, DATA		;Sends CRC_H as CRC2
	rjmp	INC_SNAP		;Move to position for T_CRC1 in pointertable



T_CRC1:	mov	DATA, CRC_L		;Sub CRC16 always called with data in DATA
	out	UDR, DATA		;Sends CRC_L as CRC1
	rjmp	INC_SNAP		;Move to position for T_END in pointertable



T_END:	rjmp	NO_SYNC






JMP2_UART_RXC_TXC_END:			;Trampoline
	rjmp	UART_RXC_TXC_END






;*** CRC16 ***
;Purpose:
;	Calculate CRC-16 Values. Uses The CCITT-16 Polynomial,
;	Expressed As X^16 + X^12 + X^5 + 1
;Author:
;	John C. Wren
;Origin:
;	http://www.kd4dts.com/download/zip/crc16.zip  [Crc16.asm]
;2000-07-17:
;	Modifyed from MCS-51 to AVR by Gran Nordahl

CRC16:	eor	CRC_H, DATA
	ldi	COUNT, 0x08
CRC16_LOOP:
	lsl	CRC_L
	rol	CRC_H
	brcc	CRC16_SKIP
	ldi	TEMP, 0x10
	eor	CRC_H, TEMP
	ldi	TEMP, 0x21
	eor	CRC_L, TEMP
CRC16_SKIP:
	dec	COUNT
	brne	CRC16_LOOP
	ret






;Pointertable. Used for Indirect Jumps. Register SNAP used as offset.
SNAP0:	.db	low(R_SYNC), high(R_SYNC)	;Offset = 0 ==> no SYNC / new packet
SNAP1:	.db	low(R_HDB2), high(R_HDB2)	;Offset = 1
SNAP2:	.db	low(R_HDB1), high(R_HDB1)	;Offset = 2
SNAP3:	.db	low(R_DAB3), high(R_DAB3)	;Offset = ...
SNAP4:	.db	low(R_DAB2), high(R_DAB2)
SNAP5:	.db	low(R_DAB1), high(R_DAB1)
SNAP6:	.db	low(R_SAB3), high(R_SAB3)
SNAP7:	.db	low(R_SAB2), high(R_SAB2)
SNAP8:	.db	low(R_SAB1), high(R_SAB1)
SNAP9:	.db	low(R_DB1),  high(R_DB1)
SNAP10:	.db	low(R_CRC2), high(R_CRC2)
SNAP11:	.db	low(R_CRC1), high(R_CRC1)	;Offset = 11 ==> end of incoming packet
SNAP12:	.db	low(T_HDB2), high(T_HDB2)
SNAP13:	.db	low(T_HDB1), high(T_HDB1)
SNAP14:	.db	low(T_DAB3), high(T_DAB3)
SNAP15:	.db	low(T_DAB2), high(T_DAB2)
SNAP16:	.db	low(T_DAB1), high(T_DAB1)
SNAP17:	.db	low(T_SAB3), high(T_SAB3)
SNAP18:	.db	low(T_SAB2), high(T_SAB2)
SNAP19:	.db	low(T_SAB1), high(T_SAB1)
SNAP20:	.db	low(T_DB1),  high(T_DB1)
SNAP21:	.db	low(T_CRC2), high(T_CRC2)
SNAP22:	.db	low(T_CRC1), high(T_CRC1)
SNAP23:	.db	low(T_END),  high(T_END)
;Vulpes^2